@@ -581,9 +581,18 @@ public abstract class KotlinBuiltIns {
 
     @NotNull
     public KotlinType getArrayElementType(@NotNull KotlinType arrayType) {
+        KotlinType result = getArrayElementTypeOrNull(arrayType);
+        if (result == null) {
+            throw new IllegalStateException("not array: " + arrayType);
+        }
+        return result;
+    }
+
+    @Nullable
+    public KotlinType getArrayElementTypeOrNull(@NotNull KotlinType arrayType) {
         if (isArray(arrayType)) {
             if (arrayType.getArguments().size() != 1) {
-                throw new IllegalStateException();
+                return null;
             }
             return arrayType.getArguments().get(0).getType();
         }
@@ -599,7 +608,7 @@ public abstract class KotlinBuiltIns {
         }
 
 
-        throw new IllegalStateException("not array: " + arrayType);
+        return null;
     }
 
     @Nullable
@@ -1004,10 +1013,6 @@ public abstract class KotlinBuiltIns {
         return classFqNameEquals(descriptor, FqNames.array);
     }
 
-    public static boolean isCloneable(@NotNull ClassDescriptor descriptor) {
-        return classFqNameEquals(descriptor, FqNames.cloneable);
-    }
-
     // This function only checks presence of Deprecated annotation at declaration-site, it doesn't take into account @DeprecatedSinceKotlin
     // To check that a referenced descriptor is actually deprecated at call-site, use DeprecationResolver
     public static boolean isDeprecated(@NotNull DeclarationDescriptor declarationDescriptor) {
